#
#    JKS - Measurement database system
#    Copyright (C) 2024  Christoph Lehner (christoph.lehner@ur.de, https://github.com/lehner/jks)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import math, random, scipy.optimize, numpy, decimal, sys
import jks.distributions


#
# performs chi2 minimization of  chi2 = dX . iCov . dX   with  dX_i = em - x_i
#
# =>  chi2 = \sum_{i,j}  (f(x_i) - x_i) iCov_{ij} (f(x_j) - x_j)
#
# In eigenrepresentation:
#
# iCov_{ij} = 1/ln vn_i vn_j => chi2 = \sum_n dx_n^2 / ln
#
# with dx_n = \sum_i dX_i vn_i = \sum_i (em - x_i) vn_i
#
# => chi2 = \sum_{n} ( \sum_i (f(x_i) - x_i) vn_i )^2 (1/ln)
#
def ad(x, eps, i):
    return [x[j] if i != j else x[j] + eps for j in range(len(x))]


class fit:

    #
    # Data type
    #
    np_datatype = numpy.float64

    def __init__(self, mean, cov, r=None):
        if r == None:
            r = range(len(mean))
        self.r = r
        self.mean = [mean[i] for i in r]
        if type(cov) != list:
            cov = [[cov]]
        self.cov = numpy.matrix([[cov[i][j] for j in r] for i in r], self.np_datatype)

        eigenvalues, eigenvectors = numpy.linalg.eig(self.cov)
        idx = eigenvalues.argsort()

        self.ev = (eigenvalues[idx], eigenvectors[:, idx])
        self.nr = range(len(self.mean))

    def chi2_compute(self, ff, eigenvalues, verbose, pars):

        f = ff([numpy.float64(p) for p in pars])

        ev = self.ev
        mean = self.mean
        nr = self.nr

        res = 0.0

        if verbose:
            print("Chi2 Computation\n------------------------------------")
            print("P    = " + str(pars))
            for i in nr:
                print("f[%d] = %g, mean[%d] = %g" % (i, f[i], i, mean[i]))
            print("")

        for n in eigenvalues:

            sp = 0.0
            for i in nr:
                sp += (f[i] - mean[i]) * ev[1][i, n]

            if verbose:
                sev = "chi2_ev%d = %g  (%g / %g)  " % (
                    n,
                    sp * sp / ev[0][n],
                    sp * sp,
                    ev[0][n],
                )
                sev += " " * (70 - len(sev))
                sev += "[ "
                for i in nr:
                    sev += "%.3g " % ev[1][i, n]
                sev += "]"
                print(sev)

            res += sp * sp / ev[0][n]

        if verbose:
            print("\n chi2 = %.15g\n" % res)

        return res

    def fit(
        self,
        f,
        scipy_method,
        _guess,
        tolerance,
        maxiter,
        verbose,
        estimate_hessians,
        freeze_index=None,
        eigenvalues=None,
    ):

        if eigenvalues is None:
            eigenvalues = self.nr

        evs = self.ev[0].tolist()

        sev = [evs[i] for i in eigenvalues]
        condition_number = sev[len(sev) - 1] / sev[0]

        if freeze_index is None:
            guess = _guess
            freeze_guess = []
        else:
            guess = _guess[0:freeze_index]
            freeze_guess = _guess[freeze_index:]

        num = 0.0
        den = 0.0
        chi2 = 0.0
        dof = len(eigenvalues) - len(guess)

        # min
        minres = scipy.optimize.minimize(
            fun=lambda pars: self.chi2_compute(
                f, eigenvalues, verbose, numpy.array(pars.tolist() + freeze_guess)
            ),
            x0=guess,
            method=scipy_method,
            tol=tolerance,
            options={"maxiter": maxiter, "disp": verbose},
        )

        guess = _guess

        if verbose and minres.success == False:
            print("Fitter failed: " + minres.message)

        if verbose:
            print("")

        hessian_PP = None
        hessian_PM = None
        if minres.success:
            val = minres.x.tolist() + freeze_guess
            chi2 = float(self.chi2_compute(f, eigenvalues, verbose, val))
            if estimate_hessians != 0.0:
                estimate_hessians = numpy.array(
                    estimate_hessians, dtype=numpy.longdouble
                )
                mean0 = self.mean
                hessian_PP = numpy.array(
                    [[0.0 for j in range(len(guess))] for i in range(len(guess))],
                    dtype=numpy.longdouble,
                )
                hessian_PM = numpy.array(
                    [[0.0 for j in range(len(mean0))] for i in range(len(guess))],
                    dtype=numpy.longdouble,
                )
                for i in range(len(guess)):
                    for j in range(len(guess)):
                        x_pp = self.chi2_compute(
                            f,
                            eigenvalues,
                            verbose,
                            ad(ad(val, estimate_hessians, j), estimate_hessians, i),
                        )
                        x_pm = self.chi2_compute(
                            f,
                            eigenvalues,
                            verbose,
                            ad(ad(val, -estimate_hessians, j), estimate_hessians, i),
                        )
                        x_mp = self.chi2_compute(
                            f,
                            eigenvalues,
                            verbose,
                            ad(ad(val, estimate_hessians, j), -estimate_hessians, i),
                        )
                        x_mm = self.chi2_compute(
                            f,
                            eigenvalues,
                            verbose,
                            ad(ad(val, -estimate_hessians, j), -estimate_hessians, i),
                        )
                        hessian_PP[i][j] = (
                            (x_pp + x_mm - x_mp - x_pm) / 4.0 / estimate_hessians**2.0
                        )
                    for j in range(len(mean0)):
                        self.mean = ad(mean0, estimate_hessians, j)
                        x_pp = self.chi2_compute(
                            f, eigenvalues, verbose, ad(val, estimate_hessians, i)
                        )
                        x_pm = self.chi2_compute(
                            f, eigenvalues, verbose, ad(val, -estimate_hessians, i)
                        )
                        self.mean = ad(mean0, -estimate_hessians, j)
                        x_mp = self.chi2_compute(
                            f, eigenvalues, verbose, ad(val, estimate_hessians, i)
                        )
                        x_mm = self.chi2_compute(
                            f, eigenvalues, verbose, ad(val, -estimate_hessians, i)
                        )
                        self.mean = mean0
                        hessian_PM[i][j] = (
                            (x_pp + x_mm - x_mp - x_pm) / 4.0 / estimate_hessians**2.0
                        )
                self.mean = mean0

        return {
            "val": minres.x.tolist(),
            "success": minres.success,
            "chi2": chi2,
            "dof": dof,
            "range": self.r,
            "p_right_tail": jks.distributions.p_value_right_tail(chi2, dof),
            "p": jks.distributions.p_value(chi2, dof),
            "eigenspace": eigenvalues,
            "condition_number": condition_number,
            "eigenvalues": evs,
            "nfev": minres.nfev,
            "hessian_PP": hessian_PP,
            "hessian_PM": hessian_PM,
        }



#
#    JKS - Measurement database system
#    Copyright (C) 2024  Christoph Lehner (christoph.lehner@ur.de, https://github.com/lehner/jks)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import scipy.special


# p-value
def p_value_right_tail(chi2, dof):
    return 1.0 - scipy.special.gammainc(dof / 2.0, chi2 / 2.0)


def p_value_left_tail(chi2, dof):
    return 1.0 - p_value_right_tail(chi2, dof)


def p_value(chi2, dof):
    # if chi2 > dof:
    return p_value_right_tail(chi2, dof)
    # else:
    #    return p_value_left_tail(chi2,dof)



#
#    JKS - Measurement database system
#    Copyright (C) 2024  Christoph Lehner (christoph.lehner@ur.de, https://github.com/lehner/jks)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import sys, os, struct, binascii, fnmatch, numpy


class writer:
    def __init__(self, fn):
        self.f = open(fn, "w+b")

    def write(self, t, cc):
        if self.f is not None:
            self.f.write(struct.pack("i", len(t) + 1))
            self.f.write((t + "\0").encode("utf-8"))
            ln = len(cc)
            ccr = [fff for sublist in ((c.real, c.imag) for c in cc) for fff in sublist]
            bindata = struct.pack("d" * 2 * ln, *ccr)
            crc32comp = binascii.crc32(bindata) & 0xFFFFFFFF
            self.f.write(struct.pack("II", crc32comp, ln))
            self.f.write(bindata)
            self.f.flush()

    def close(self):
        if self.f is not None:
            self.f.close()
            self.f = None


def count(fn):
    n = 0
    f = open(fn, "rb")
    while True:
        rd = f.read(4)
        if len(rd) == 0:
            break
        ntag = struct.unpack("i", rd)[0]
        f.read(ntag)
        (crc32, ln) = struct.unpack("II", f.read(4 * 2))

        f.read(16 * ln)
        n += 1
    f.close()
    return n


def tags(fn, filter=lambda x: x):
    n = 0
    tags = []
    f = open(fn, "rb")
    while True:
        rd = f.read(4)
        if len(rd) == 0:
            break
        ntag = struct.unpack("i", rd)[0]
        tag = f.read(ntag).decode("utf-8")[0:-1]
        (crc32, ln) = struct.unpack("II", f.read(4 * 2))

        data = f.read(16 * ln)

        crc32comp = binascii.crc32(data) & 0xFFFFFFFF

        if crc32comp != crc32:
            raise Exception("Data corrupted!")

        corr = numpy.frombuffer(data, dtype=numpy.complex128, count=ln)

        x = filter(tag, corr)
        if x is not None:
            tags.append(x)
    f.close()
    return tags


class reader:
    def __init__(self, fn):
        self.tags = {}
        self.ntags = {}
        f = open(fn, "r+b")
        while True:
            rd = f.read(4)
            if len(rd) == 0:
                break
            ntag = struct.unpack("i", rd)[0]
            tag = f.read(ntag).decode("utf-8")
            (crc32, ln) = struct.unpack("II", f.read(4 * 2))

            data = f.read(16 * ln)
            crc32comp = binascii.crc32(data) & 0xFFFFFFFF

            if crc32comp != crc32:
                raise Exception("Data corrupted!")

            nt = tag[0:-1]
            self.tags[nt] = numpy.frombuffer(data, dtype=numpy.complex128, count=ln)
            if nt not in self.ntags:
                self.ntags[nt] = 1
            else:
                self.ntags[nt] += 1
        f.close()

    def glob(self, pattern):
        return filter(lambda k: fnmatch.fnmatch(k, pattern), self.tags.keys())



#
#    JKS - Measurement database system
#    Copyright (C) 2024  Christoph Lehner (christoph.lehner@ur.de, https://github.com/lehner/jks)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import lz4.frame

# import h5py
import math
import random
import numpy
import datetime
import time
import socket
import sys
import os
import pickle
import fnmatch
from jks.measurements import measurements

t0 = 0
t1 = 0
t2 = 0


# Resamples
class resamples:
    def __init__(self, fn=None):
        self.set = {}
        self.N = None
        self.tags = None
        self.info = None
        if fn != None:
            self.load(fn)

    def add(self, n, s):
        if self.N == None:
            self.N = s.N
            self.tags = s.tags
            self.info = s.info
            self._clone_type_str = s._clone_type_str
            self.set[n] = s
        else:
            assert self._clone_type_str == s._clone_type_str
            all_tags = sorted(set(self.tags) | set(s.tags))

            # Adding more measurements of existing n not possible beyond
            # primary data.
            # How do we go from f(1/n sum_{i=1}^n x_i) and f(1/m sum_{i=1}^m y_i)
            # to f(1/(n+m) (\sum_i x_i + \sum_j y_j)) ?
            assert n not in self.set

            # Do I need to update existing tags to make room for new tags required by s?
            update_existing_tags = all_tags != self.tags

            self.N = len(all_tags)
            self.tags = all_tags

            for t in s.info:
                if t not in self.info:
                    self.info[t] = s.info[t]
                elif self.info[t] != s.info[t]:
                    # print("Merge ", self.info[t], s.info[t])
                    if len(self.info[t]) > 1024:
                        self.info[t] = self.info[t][0:1024] + "..."
                    self.info[t] = self.info[t] + "\n" + s.info[t]
            s.info = self.info

            # Enlarge all existing sets if needed
            if update_existing_tags:
                for nn in self.set:
                    if self.set[nn].tags != self.tags:
                        self.set[nn].expand(self.tags)

            # Enlarge to-be-added
            if s.tags != self.tags:
                s.expand(self.tags)

            self.set[n] = s

    def add_gaussian(self, n, m, e):
        fnc = self._clone_type_fnc()
        # TODO: add new tags
        self.set[n] = fnc(
            m, [random.gauss(m, e / math.sqrt(self.N - 1)) for i in range(self.N)]
        )

    def cut(self, i, keep_fixed=None):
        if i == -1:
            return dict([(t, self.set[t].orig) for t in self.set.keys()])
        else:
            if keep_fixed != None:
                return dict(
                    [
                        (
                            t,
                            (
                                self.set[t].blocks[i]
                                if t not in keep_fixed
                                else self.set[t].orig
                            ),
                        )
                        for t in self.set.keys()
                    ]
                )
            else:
                return dict([(t, self.set[t].blocks[i]) for t in self.set.keys()])

    def _clone_type_fnc(self):
        return eval(
            "lambda a,b,c,d: "
            + self._clone_type_str
            + "({ 'orig' : a, 'blocks' : b, 'tags' : c, 'info' : d })"
        )

    def apply(self, phi, keep_fixed=None, scale=1.0):
        f = self._clone_type_fnc()
        if scale == 1.0:
            return f(
                phi(self.cut(-1)),
                [phi(self.cut(i, keep_fixed)) for i in range(self.N)],
                self.tags,
                self.info,
            )
        else:
            x = self.cut(-1)
            y = numpy.array(phi(x))
            yi0 = [
                phi(
                    dict(
                        [
                            (
                                t,
                                (numpy.array(xi) - numpy.array(x[t])) * scale
                                + numpy.array(x[t]),
                            )
                            for t, xi in self.cut(i, keep_fixed).items()
                        ]
                    )
                )
                for i in range(self.N)
            ]
            yi = [(numpy.array(yi0[i]) - y) / scale + y for i in range(self.N)]
            return f(y, yi, self.tags, self.info)

    def get(self, n):
        return self.set[n]

    def take(self, other, pat):
        for n in other.set:
            if fnmatch.fnmatch(n, pat):
                self.add(n, other.get(n))

    def keys(self):
        return self.set.keys()

    def match(self, pat):
        return [k for k in self.set if fnmatch.fnmatch(k, pat)]

    def save(self, fn, compress=True):

        t0 = time.time()
        s = {
            "N": self.N,
            "set": {},
            "_clone_type_str": self._clone_type_str,
            "tags": self.tags,
            "info": self.info,
            "origin": {
                "argv": sys.argv,
                "pwd": os.getcwd(),
                "hostname": socket.gethostname(),
                "user": os.getlogin(),
                "date": str(datetime.datetime.now()),
                "argv[0].mtime": str(
                    datetime.datetime.fromtimestamp(os.path.getmtime(sys.argv[0]))
                ),
            },
        }
        for f in self.set:
            s["set"][f] = {
                "orig": self.set[f].orig,
                "blocks": self.set[f].blocks,
                "N": self.set[f].N,
            }

        t1 = time.time()

        if compress:
            with lz4.frame.open(fn, mode="wb") as f:
                pickle.dump(s, f, pickle.HIGHEST_PROTOCOL)
        else:
            with open(fn, "wb") as f:
                pickle.dump(s, f, pickle.HIGHEST_PROTOCOL)
        # jksIO.save(fn,s)
        t2 = time.time()
        # print "Time: %g(prep) %g(wr)" % (t1-t0,t2-t1)

    def load(self, fn, pre=""):
        assert self.N == None
        try:
            with lz4.frame.open(fn, mode="rb") as f:
                s = pickle.load(f, encoding="latin1", errors="")
        except:
            with open(fn, "rb") as f:
                s = pickle.load(f, encoding="latin1", errors="")
        #
        self.N = s["N"]
        if "tags" in s:
            self.tags = s["tags"]
        else:
            # old versions did not save tags
            self.tags = ["n/a-%6.6d" % i for i in range(self.N)]
        if "info" in s:
            self.info = s["info"]
        else:
            self.info = {}
        self._clone_type_str = s["_clone_type_str"]
        fnc = self._clone_type_fnc()
        for f in s["set"]:
            self.set[pre + f] = fnc(
                s["set"][f]["orig"], s["set"][f]["blocks"], self.tags, self.info
            )

    def used(self):
        used = set([])
        for f in self.set:
            o = numpy.nan_to_num(self.set[f].orig)
            b = numpy.nan_to_num(self.set[f].blocks)
            for i in range(len(b)):
                if not numpy.array_equal(o, b[i]):
                    used.add(i)
        return list(used)

    def compress(self, verbose=False):
        used = sorted(self.used())
        N = len(used)
        for i in range(len(self.tags)):
            if i not in used:
                if self.tags[i][0] == "!":
                    t = self.tags[i][1:]
                    del self.info[t]
                    if verbose:
                        print("Delete variation", t)
                else:
                    if verbose:
                        print("Delete config", self.tags[i])
        self.tags = [self.tags[i] for i in used]
        self.N = N
        for f in self.set:
            self.set[f].blocks = [self.set[f].blocks[i] for i in used]
            self.set[f].N = N
            self.set[f].tags = self.tags

    def prepend_tags(self, pre):
        self.tags = [pre + t if t[0] != "!" else t for t in self.tags]

    def summary(self, phi):
        rss = self.apply(phi)
        mn = rss.mean()
        cv0 = rss.cov()
        res = {}
        for k in self.keys():
            rss = self.apply(phi, [k])
            cvk = rss.cov()
            if cvk != cv0:
                res[cvk] = k
        sk = sorted(res.keys())
        print("----------------------------------")
        print(" Summary of error contributions ")
        print("----------------------------------")
        print(" - %g original" % math.sqrt(cv0))
        for s in sk:
            print(" - %g for %s fixed" % (math.sqrt(s), res[s]))
        print("----------------------------------")



#
#    JKS - Measurement database system
#    Copyright (C) 2024  Christoph Lehner (christoph.lehner@ur.de, https://github.com/lehner/jks)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import math, numpy
from jks.resamples import resamples
from jks.measurements import measurements, jackknife
from jks.fit import fit
import jks.corrIO


def qfit(
    mn,
    cv,
    r,
    fnc,
    guess,
    tolerance=1e-7,
    maxiter=100,
    verbose=False,
    simplex=True,
    estimate_hessians=0.0,
    freeze_index=None,
):
    return fit(mn, cv, r).fit(
        lambda pars: [fnc(i, pars) for i in r],
        "Nelder-Mead",
        guess,
        tolerance,
        maxiter,
        verbose,
        estimate_hessians,
        freeze_index,
    )


# write helper function
def write(fn, arr):
    f = open(fn, "w")
    for row in arr:
        for col in row:
            f.write(str(col))
            f.write(" ")
        f.write("\n")
    f.close()


# format value and error
def gformat(value, errors, etags, times="\\times"):
    err_digits = 2
    error = max([math.fabs(errors[e]) for e in etags])
    val_digits_left_of_period = int(
        math.floor(1e-10 + math.log(math.fabs(value)) / math.log(10.0)) + 1
    )
    if error == 0.0:
        return "%.15g" % (value)
    else:
        err_digits_left_of_period = int(
            math.floor(1e-10 + math.log(error) / math.log(10.0)) + 1
        )

    digits_left_of_period = max([val_digits_left_of_period, err_digits_left_of_period])
    digits_right_of_period = err_digits_left_of_period - err_digits
    if digits_left_of_period < 0:
        digits_shift = -digits_left_of_period + 1
    elif digits_right_of_period > 0:
        digits_shift = -digits_right_of_period
    else:
        digits_shift = 0
    value *= 10**digits_shift
    errors_rescaled = dict(
        [(e, errors[e] * 10**-digits_right_of_period) for e in etags]
    )
    digits_left_of_period += digits_shift
    digits_right_of_period += digits_shift

    s = ("{0:%d.%df}" % (digits_left_of_period, -digits_right_of_period)).format(value)
    for e in etags:
        if errors[e] == 0.0:
            continue
        if digits_right_of_period == -1:
            s = s + "(" + "{0:02.1f}".format(errors_rescaled[e] / 10.0) + ")"
        else:
            s = s + "(" + "{0:02.0f}".format(errors_rescaled[e]) + ")"
        if e != "":
            s = s + "_{%s}" % e
    if digits_shift != 0:
        if digits_shift == -1:
            s += " %s 10" % times
        else:
            s += " %s 10^{%d}" % (times, -digits_shift)
    return s


def format(value, error, error_ndigits=None):
    exp_digits = 0
    err_extr_digits = 0
    if error_ndigits != None and error_ndigits < 0:
        err_extr_digits = -error_ndigits
        error_ndigits = None
    if error_ndigits == None:
        error_ndigits = (
            int(math.floor(-math.log(error) / math.log(10))) + 1 + err_extr_digits
        )
        if error_ndigits < 0:
            error *= 10.0**error_ndigits
            value *= 10.0**error_ndigits
            exp_digits = -error_ndigits
            error_ndigits = 0
    value_rescaled = round(value * math.pow(10, error_ndigits)) * math.pow(
        10, -error_ndigits
    )
    error_rescaled = int(round(error * math.pow(10, error_ndigits)))
    s = (
        ("{:." + str(error_ndigits) + "f}").format(value_rescaled)
        + "("
        + str(error_rescaled)
        + ")"
    )
    if exp_digits > 0:
        if exp_digits == 1:
            s += " x 10"
        else:
            s += " x 10^%d" % exp_digits
    return s


#
def write_confidence_band(f, pars, pars_cov, eps, xrang, outfn):
    # f(x,pars) = f(x,pars) +- sqrt( (df/dpars_i) (x,pars) pars_cov_ij (df/dpars_j) (x,pars) )
    fout = open(outfn, "wt")
    for x in xrang:
        fx = f(x, pars)
        dfx = [
            (
                f(x, [pars[i] + eps if i == j else pars[i] for i in range(len(pars))])
                - f(x, pars)
            )
            / eps
            for j in range(len(pars))
        ]
        var = 0.0
        for i in range(len(pars)):
            for j in range(len(pars)):
                var += dfx[i] * dfx[j] * pars_cov[i][j]

        if type(x) == type(0.0):
            pos = "%g" % x
        elif type(x) == type(0):
            pos = "%d" % x
        elif type(x) == type([]):
            pos = str(x)

        fout.write("%s %.15g %.15g\n" % (pos, fx, math.sqrt(var)))

    fout.close()


def gaussian_noise(cov, n):
    r = range(len(cov))
    np_datatype = numpy.float64
    cov = numpy.matrix([[cov[i][j] for j in r] for i in r], np_datatype)

    # check if symmetric real matrix
    assert numpy.isrealobj(cov)
    assert numpy.allclose(cov, cov.T)

    eigenvalues, eigenvectors = numpy.linalg.eigh(cov)
    idx = eigenvalues.argsort()

    evals, evecs = eigenvalues[idx], eigenvectors[:, idx].tolist()
    largest = sorted(evals)[-1]
    evals[numpy.abs(evals) < 1e-15 * largest] = (1e-100) * largest
    evals = evals.tolist()
    evr = [
        numpy.random.normal(0.0, math.sqrt(evals[i]), n).tolist()
        for i in range(len(evals))
    ]
    return [[sum([evr[i][j] * evecs[l][i] for i in r]) for l in r] for j in range(n)]


def draw(m, cov, n):
    es = gaussian_noise(cov, n)
    return [[m[i] + e[i] for i in range(len(m))] for e in es]



from setuptools import setup, find_packages
import glob

scripts = glob.glob("scripts/*")

setup(
    name="jks-system",
    version="1.0.2",
    packages=find_packages(),  # Automatically find the packages in the project
    scripts=scripts,  # Include scripts as is
    entry_points={},
    install_requires=["scipy","lz4","numpy"],  # List your package dependencies here
    author="Christoph Lehner",
    author_email="christoph@lhnr.de",
    description="JKS Measurement database system",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/lehner/jks",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
